from grai_schemas import (
    base,
    generics,
    human_ids,
    integrations,
    package_definitions,
    schema,
    serializers,
    utilities,
    v1,
)
from grai_schemas.package_definitions import config

__version__ = "0.2.7-alpha3"



from abc import ABC, abstractmethod
from typing import List, Literal, Set

from grai_schemas.base import Event, SourcedEdge, SourcedNode
from pydantic import BaseModel


class QuarantineReason(BaseModel, ABC):
    @property
    @abstractmethod
    def reason(self) -> str:
        pass


class MissingEdgeNodeReason(QuarantineReason):
    side: Literal["source", "destination"]
    node_name: str
    node_namespace: str

    @property
    def reason(self) -> str:
        return f"{self.side.capitalize()} node `({self.node_name}, {self.node_namespace})` was not found."


class QuarantinedEdge(BaseModel):
    edge: SourcedEdge
    reasons: List[QuarantineReason]


class QuarantinedNode(BaseModel):
    node: SourcedNode
    reasons: List[QuarantineReason]


class QuarantinedEvent(BaseModel):
    event: Event
    reasons: List[QuarantineReason]


class Quarantine(BaseModel):
    nodes: List[QuarantinedNode] = []
    edges: List[QuarantinedEdge] = []
    events: List[QuarantinedEvent] = []

    @property
    def has_quarantined(self):
        return self.nodes or self.edges or self.events



from __future__ import annotations

import uuid
from typing import TYPE_CHECKING, Any, Dict, Generic, List, Tuple, TypeVar, Union

from multimethod import multimethod
from pydantic import BaseModel

T = TypeVar("T")


def unpack_object(obj: Union[Dict, BaseModel]) -> Dict:
    """

    Args:
        obj (Union[Dict, BaseModel]):

    Returns:

    Raises:

    """
    if isinstance(obj, Dict):
        return obj
    elif isinstance(obj, BaseModel):
        return obj.dict()
    else:
        raise NotImplementedError(f"No method to unpack objects of type {type(obj)}")


atomic = Union[int, float, complex, str, bool, uuid.UUID]


@multimethod
def merge(a, b):
    """

    Args:
        a:
        b:

    Returns:

    Raises:

    """
    raise Exception(f"No merge logic for type {type(a)} and {type(b)}")


@merge.register
def merge_atomic(a: Any, b: T) -> T:
    """

    Args:
        a (Any):
        b (T):

    Returns:

    Raises:

    """
    return b


@merge.register
def merge_missing(a: T, b: None) -> T:
    """

    Args:
        a (T):
        b (None):

    Returns:

    Raises:

    """
    return a


@merge.register
def merge_dict_item(a: Dict, b: Dict) -> Dict:
    """

    Args:
        a (Dict):
        b (Dict):

    Returns:

    Raises:

    """
    result = {**a, **b}
    result.update({key: merge(a[key], b[key]) for key in a.keys() & b.keys()})
    return result


@merge.register
def merge_list(a: list, b: list) -> list:
    """

    Args:
        a (list):
        b (list):

    Returns:

    Raises:

    """
    return [*a, *b]


@merge.register
def merge_tuple(a: tuple, b: tuple) -> tuple:
    """

    Args:
        a (tuple):
        b (tuple):

    Returns:

    Raises:

    """
    return *a, *b


@merge.register
def merge_set(a: set, b: set) -> set:
    """

    Args:
        a (set):
        b (set):

    Returns:

    Raises:

    """
    return b | a


@merge.register
def merge_pydantic(a: BaseModel, b: Any) -> BaseModel:
    """

    Args:
        a (BaseModel):
        b (Any):

    Returns:

    Raises:

    """
    merged = merge(dict(a), b)
    return type(a)(**merged)


@merge.register
def merge_pydantic_right(a: T, b: BaseModel) -> T:
    """

    Args:
        a (T):
        b (BaseModel):

    Returns:

    Raises:

    """
    return merge(a, dict(b))


def merge_models(a: T, b: T) -> T:
    """This function is deprecated. Use `merge` instead

    Args:
        a (T):
        b (T):

    Returns:

    Raises:

    """
    a_type = type(a)
    merged = merge(a, b)
    return a_type(**merged)


def compute_graph_changes(items: List[T], active_items: List[T]) -> Tuple[List[T], List[T], List[T]]:
    """
    Args:
        items:
        active_items:

    Returns:

    Raises:

    """

    active_item_map = {hash(item.spec): item for item in active_items}
    item_map: Dict[int, T] = {hash(item.spec): item for item in items}

    new_item_keys = item_map.keys() - active_item_map.keys()
    deleted_source_item_keys = active_item_map.keys() - item_map.keys()
    updated_item_keys = item_map.keys() - new_item_keys

    deleted_from_sources = [active_item_map[k] for k in deleted_source_item_keys]

    new_items: List[T] = [item_map[k] for k in new_item_keys]
    updated_items = [item_map[k] for k in updated_item_keys if item_map[k] != active_item_map[k]]

    return new_items, updated_items, deleted_from_sources



import sys
from abc import ABC, abstractmethod
from typing import List, Optional, Tuple, Union

from grai_schemas.base import Event, SourcedEdge, SourcedNode
from grai_schemas.integrations.quarantine import (
    MissingEdgeNodeReason,
    Quarantine,
    QuarantinedEdge,
    QuarantinedEvent,
    QuarantinedNode,
)
from grai_schemas.v1.source import SourceSpec, SourceV1

if sys.version_info < (3, 10):
    from typing_extensions import ParamSpec
else:
    from typing import ParamSpec

from functools import cache

from pydantic import BaseModel

P = ParamSpec("P")


def verify_edge_ids(
    nodes: List[SourcedNode], edges: List[SourcedEdge]
) -> Tuple[List[SourcedEdge], List[QuarantinedEdge]]:
    node_labels = {(n.spec.namespace, n.spec.name) for n in nodes}

    good_edges = []
    quarantined_edges = []
    for edge in edges:
        source_id = (edge.spec.source.namespace, edge.spec.source.name)
        destination_id = (edge.spec.destination.namespace, edge.spec.destination.name)
        if source_id in node_labels and destination_id in node_labels:
            good_edges.append(edge)
        else:
            reasons = [
                MissingEdgeNodeReason(side=side, node_name=name, node_namespace=namespace)
                for side, (namespace, name) in (("source", source_id), ("destination", destination_id))
                if (namespace, name) not in node_labels
            ]
            quarantined_edges.append(QuarantinedEdge(edge=edge, reasons=reasons))

    # TODO: sentry alerting on quarantine?

    return good_edges, quarantined_edges


class ValidatedResult(BaseModel):
    nodes: List[SourcedNode] = []
    edges: List[SourcedEdge] = []
    events: List[Event] = []


class GraiIntegrationImplementation(ABC):
    source: SourceV1
    version: str

    def __init__(
        self,
        source: SourceV1,
        version: Optional[str] = None,
    ):
        self.source = source
        self.version = version if version else "v1"

    @abstractmethod
    def nodes(self) -> List[SourcedNode]:
        pass

    @abstractmethod
    def edges(self) -> List[SourcedEdge]:
        pass

    @abstractmethod
    def get_nodes_and_edges(self) -> Tuple[List[SourcedNode], List[SourcedEdge]]:
        pass

    @abstractmethod
    def ready(self) -> bool:
        pass

    def events(self, *args, **kwargs) -> List[Event]:
        return []


class QuarantineAccessor:
    def __init__(self, integration_instance: GraiIntegrationImplementation):
        self._integration = integration_instance
        self._quarantine = Quarantine()

    @property
    def nodes(self) -> List[QuarantinedNode]:
        _ = self._integration.nodes  # Ensure nodes are populated in the quarantine
        return self._quarantine.nodes

    @nodes.setter
    def nodes(self, value: List[QuarantinedNode]):
        self._quarantine.nodes = value

    @property
    def edges(self) -> List[QuarantinedEdge]:
        _ = self._integration.edges  # Ensure edges are populated in the quarantine
        return self._quarantine.edges

    @edges.setter
    def edges(self, value: List[QuarantinedEdge]):
        self._quarantine.edges = value

    @property
    def events(self) -> List[QuarantinedEvent]:
        _ = self._integration.events  # Ensure events are populated in the quarantine
        return self._quarantine.events

    @events.setter
    def events(self, value: List[QuarantinedEvent]):
        self._quarantine.events = value

    @property
    def has_quarantined(self) -> bool:
        return self._quarantine.has_quarantined


class ValidatedIntegration(GraiIntegrationImplementation):
    def __init__(self, integration: GraiIntegrationImplementation, *args, **kwargs):
        self.integration = integration
        self._quarantine_accessor = QuarantineAccessor(self)
        super().__init__(source=self.integration.source, version=self.integration.version)

    @property
    def quarantine(self) -> QuarantineAccessor:
        return self._quarantine_accessor

    @cache
    def nodes(self) -> List[SourcedNode]:
        return self.integration.nodes()

    @cache
    def edges(self) -> List[SourcedEdge]:
        edges, quarantined_edges = verify_edge_ids(*self.integration.get_nodes_and_edges())
        self.quarantine.edges = quarantined_edges
        return edges

    @cache
    def get_nodes_and_edges(self) -> Tuple[List[SourcedNode], List[SourcedEdge]]:
        return self.nodes(), self.edges()

    @cache
    def events(self, *args, **kwargs) -> List[Event]:
        return self.integration.events(*args, **kwargs)

    def ready(self) -> bool:
        return self.integration.ready()
